﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Data.Entity;
using System.ServiceModel.DomainServices.Server;
using System.Web;

namespace BigShelf.Models
{
    public partial class BigShelfEntities : DbContext
    {
        protected override void OnModelCreating(DbModelBuilder modelBuilder)
        {
            base.OnModelCreating(modelBuilder);
            modelBuilder.Entity<Friend>()
                        .HasRequired(e => e.Profile)
                        .WithMany(e => e.Friends)
                        .HasForeignKey(e => e.ProfileId)
                        .WillCascadeOnDelete(false);
        }

        public BigShelfEntities()
            : base("name=BigShelfEntities")
        {
            if (HttpContext.Current == null)
            {
                Database.SetInitializer<BigShelfEntities>(null);
            }
            else
            {
                Database.SetInitializer<BigShelfEntities>(new BigShelfContextInitializer());
            }
        }

        public DbSet<Book> Books { get; set; }

        public DbSet<Category> Categories { get; set; }

        public DbSet<CategoryName> CategoryNames { get; set; }

        public DbSet<FlaggedBook> FlaggedBooks { get; set; }

        public DbSet<Friend> Friends { get; set; }

        public DbSet<Profile> Profiles { get; set; }
    }

    public class BigShelfContextInitializer : DropCreateDatabaseAlways<BigShelfEntities>
    {
        protected override void Seed(BigShelfEntities context)
        {
            var profiles = new List<Profile>
                {
                    new Profile { Id = 1, Name = "Demo User", EmailAddress = "demo@microsoft.com", AspNetUserGuid = "3730F12D-1A2A-4499-859B-9F586B2858A4" },
                    new Profile { Id = 2, Name = "Deepesh Mohnani", EmailAddress = "deepm@microsoft.com", AspNetUserGuid = "5ad7976c-7a95-47aa-87ba-29c3fc80643e" },
                    new Profile { Id = 3, Name = "Jeff Handley", EmailAddress = "jeffhand@microsoft.com", AspNetUserGuid = "9330619d-9a8c-4269-9bde-ba4cb1b7b354" },
                    new Profile { Id = 4, Name = "Yavor Georgiev", EmailAddress = "yavorg@microsoft.com", AspNetUserGuid = "1a481f1e-1b22-4ae3-a75d-d5fa664d4085" },
                    new Profile { Id = 5, Name = "Howard Dierking", EmailAddress = "howard@microsoft.com", AspNetUserGuid = "5df20314-4c44-44ad-9847-143f28f85bf5" },
                    new Profile { Id = 6, Name = "Glenn Block", EmailAddress = "gblock@microsoft.com", AspNetUserGuid = "990c62c5-30a4-4ca9-92c5-f248241f6d44" },
                    new Profile { Id = 7, Name = "Joseph Connolly", EmailAddress = "joconnol@microsoft.com", AspNetUserGuid = "3e82b681-81ce-4c87-856a-61381d4ad0e8" },
                    new Profile { Id = 8, Name = "Brad Olenick", EmailAddress = "brado@microsoft.com", AspNetUserGuid = "9fc20e29-91b1-47c0-8054-5624ac87c5d8" }
                };

            profiles.ForEach(p => context.Profiles.Add(p));

            var friends = new List<Friend>{
                new Friend {Id = 1, Profile = profiles[0], FriendProfile = profiles[1]},
                new Friend {Id = 2, Profile = profiles[0], FriendProfile = profiles[2]},
                new Friend {Id = 3, Profile = profiles[2], FriendProfile = profiles[1]},
                new Friend {Id = 4, Profile = profiles[0], FriendProfile = profiles[5]}
            };

            friends.ForEach(f => context.Friends.Add(f));

            var categoryNames = new List<CategoryName>{
                new CategoryName { Id = 1, Name = "Technology"},
                new CategoryName { Id = 2, Name = "Business"},
                new CategoryName { Id = 3, Name = "Travel"},
                new CategoryName { Id = 4, Name = "General"}
            };

            categoryNames.ForEach(cn => context.CategoryNames.Add(cn));

            var categories = new List<Category>(){
                new Category() { Id = 1, Profile = profiles[0], CategoryName = categoryNames[0]},
                new Category() { Id = 2, Profile = profiles[0], CategoryName = categoryNames[1]}
            };

            categories.ForEach(c => context.Categories.Add(c));

            var books = new List<Book>{
                new Book{ Id = 1,	CategoryName = categoryNames[0], ASIN = "0596523092", PublishDate = DateTime.Parse("2008-12-22"),  AddedDate = DateTime.Parse("2009-01-22"), Title = "Data-Driven Services with Silverlight 2"                                                                                         ,        Author = "John Papa"                  ,  Description = "This comprehensive book teaches you how to build data-rich business applications with Silverlight 2 that draw on multiple sources of data. Packed with reusable examples, Data-Driven Services with Silverlight 2 covers all of the data access and web service tools you need, including data binding, the LINQ data querying component, RESTful and SOAP web service calls, and Microsoft's new ADO.NET Data Services and the ADO.NET Framework. With this book, you will:     Know when and how to use LINQ to JSON, LINQ to XML, and LINQ to Objects  Learn how Silverlight 2 applications bind, pass, read, save, query and present data   Discover how your application can call web services to work with SOAP, REST, RSS, AtomPub, POX and JSON Design REST, ASMX, and WCF web services that communicate with Silverlight 2  Harness RESTful web services such as Digg, Amazon and Twitter  Retrieve and save data using the new Entity Framework and WCF  Work with RESTful ADO.NET Data Services and its Silverlight client library to move data between your Silverlight application and database     Data-Driven Services with Silverlight 2 offers many tips and tricks for building data-rich business applications, and covers the scenarios you're most likely to encounter. Complete examples in C# and VB can be downloaded from the books companion website.&quot;" },
                new Book{ Id = 2,	CategoryName = categoryNames[0], ASIN = "059652028X", PublishDate = DateTime.Parse("2009-02-03"),  AddedDate = DateTime.Parse("2009-03-03"), Title = "Programming Entity Framework"                                                                                                    ,        Author = "Julia Lerman"               ,  Description = "Written by Julia Lerman, the leading independent authority on the framework, Programming Entity Framework includes scores of reusable examples--written in both Visual Basic and C#--that you can implement right away. This book will help you:      Understand the core concepts you need to make the best use of the Entity Framework (EF) in your applications   Learn to query your data, using either LINQ to Entities or Entity SQL   Create Windows Forms, WPF, and ASP.NET applications   Build ASMX web services and WCF services   Use Object Services to work directly with your entity objects  Delve into model customization, relationship management, change tracking, data concurrency, and more    Presented in a clear narrative style that reflects the hundreds of hours the author has spent consulting, teaching, and writing about this new data access technology and testing its myriad features, Programming Entity Framework will help you master the technology and put it to work.&quot;" },
                new Book{ Id = 3,	CategoryName = categoryNames[0], ASIN = "0321554167", PublishDate = DateTime.Parse("2009-10-01"),  AddedDate = DateTime.Parse("2009-11-01"), Title = "Essential Silverlight 3"                                                                                                         ,        Author = "Ashraf Michail"             ,  Description = "&quot;The book you are holding contains the keys to writing great Silverlight 3 applications. Use them well!&quot; From the Foreword by Scott Guthrie, corporate vice president, Microsoft Developer Division  .  Microsoft Silverlight is a leading cross-platform, Rich Internet Application (RIA) technology that allows you to create more compelling and interactive Web experiences than are possible with simple HTML browser pages. Silverlight 3 includes significant improvements over previous versions, including out-of-browser operation, H.264 video, 3D graphics, and real-time effec.  Essential Silverlight 3 is the definitive reference and insider's guide. It not only covers all the key features of the Silverlight 3 runtime and how to use them but, in Under the Hood sections, explains why each feature was developed and how each one works. These â€œinsiderâ€ explanations often lead to concise, practical performance tips that can help you speed up your own Silverlight applications. Author Ashraf Michail is uniquely qualified to explain Silverlight 3. He is a Microsoft Silverlight architect who has guided Silverlight from its beginnings through the current version. In this book, he shows how to.      Integrate the principles and components of a Silverlight application, including XAP, XAML, and .NET code    Use Silverlight vector graphics, bitmap images, and animation    Display predictable and readable text across platforms and animate text elements    Obtain mouse and keyboard input, including mouse wheel events    Implement smooth animations to improve your user interface and make your application more pleasing to the user    Use the Canvas, StackPanel, Grid, and Border elements to lay out your application and define custom layout elements    Play high-quality, live, and on-demand video and audio    Customize the built-in and toolkit controls and develop new custom Silverlight controls    Connect to and synchronize with data using Silverlight Data Binding    Use built-in real-time effects and define custom HLSL pixel shaders    Enable GPU-accelerated rendering.  If you are a developer who is getting started with Silverlight or an expert Silverlight developer who is interested in understanding the inner workings of the Silverlight runtime, this book is for you.&quot;" },
                new Book{ Id = 4,	CategoryName = categoryNames[0], ASIN = "0321564162", PublishDate = DateTime.Parse("2009-03-22"),  AddedDate = DateTime.Parse("2009-04-22"), Title = "Essential LINQ"                                                                                                                  ,        Author = "Charlie Calvert"            ,  Description = "Charlie and Dinesh bring important skills to this project that enable them to show how LINQ works and the practical ways you can use it in your daily development process.â€  From the Foreword by Anders Hejlsberg. LINQ is one of Microsoft's most exciting, powerful new development technologies. Essential LINQ is the first LINQ book written by leading members of Microsoft's LINQ and C# teams. Writing for architects, developers, and development managers, these Microsoft insiders share their intimate understanding of LINQ, revealing new patterns and best practices for getting the most out of it. Calvert and Kulkarni begin by clearly explaining how LINQ resolves the long-time â€œimpedance mismatchâ€ between object-oriented code and relational databases. Next, they show how LINQ integrates querying into C# as a â€œfirst-class citizen.â€ Using realistic code examples, they show how LINQ provides a strongly typed, IntelliSense-aware technology for working with data from any source, including SQL databases, XML files, and generic data structures. Calvert and Kulkarni carefully explain LINQ's transformative, composable, and declarative capabilities. By fully illuminating these three concepts, the authors allow developers to discover LINQ's full power. In addition to covering core concepts and hands-on LINQ development in C# with LINQ to Objects, LINQ to XML, LINQ to SQL, and LINQ to Entities, they also present advanced topics and new LINQ implementations developed by the LINQ community.&quot;" },
                new Book{ Id = 5,	CategoryName = categoryNames[0], ASIN = "0321545613", PublishDate = DateTime.Parse("2008-11-01"),  AddedDate = DateTime.Parse("2008-12-01"), Title = "Framework Design Guidelines: Conventions, Idioms, and Patterns for Reusable .NET Libraries (2nd Edition)"                        ,        Author = "Krzysztof Cwalina"          ,  Description = "Framework Design Guidelines, Second Edition, teaches developers the best practices for designing reusable libraries for the Microsoft .NET Framework. Expanded and updated for .NET 3.5, this new edition focuses on the design issues that directly affect the programmability of a class library, specifically its publicly accessible APIs.  .  This book can improve the work of any .NET developer producing code that other developers will use. It includes copious annotations to the guidelines by thirty-five prominent architects and practitioners of the .NET Framework, providing a lively discussion of the reasons for the guidelines as well as examples of when to break those guidelines.&quot;" },
                new Book{ Id = 6,	CategoryName = categoryNames[0], ASIN = "0321562992", PublishDate = DateTime.Parse("2008-10-18"),  AddedDate = DateTime.Parse("2008-11-18"), Title = "The C# Programming Language (3rd Edition)"                                                                                       ,        Author = "Anders Hejlsberg"           ,  Description = "Based on my own experience, I can safely say that every .NET developer who reads this will have at least one aha moment and will be a better developer for it. The popular C# programming language combines the high productivity of rapid application development languages with the raw power of C and C++. Now, C# 3.0 adds functional programming techniques and LINQ, Language Integrated Query. The C# Programming Language, Third Edition, is the authoritative and annotated technical reference for C# 3.0. Written by Anders Hejlsberg, the language's architect, and his colleagues, Mads Torgersen, Scott Wiltamuth, and Peter Golde, this volume has been completely updated and reorganized for C# 3.0. The book provides the complete specification of the language, along with descriptions, reference materials, code samples, and annotations from nine prominent C# gurus. The many annotations a new feature in this edition bring a depth and breadth of understanding rarely found in any programming book. As the main text of the book introduces the concepts of the C# language, cogent annotations explain why they are important, how they are used, how they relate to other languages, and even how they evolved. This book is the definitive, must-have reference for any developer who wants to understand C#.&quot;" },
                new Book{ Id = 7,	CategoryName = categoryNames[0], ASIN = "0071597832", PublishDate = DateTime.Parse("2008-12-19"),  AddedDate = DateTime.Parse("2009-01-19"), Title = "LINQ Programming"                                                                                                                ,        Author = "Joe Mayo"                   ,  Description = "Take your LINQ programming skills to the top tierResolve data format impedance mismatch with guidance from a C# expert and Microsoft MVP. LINQ Programming details cutting-edge techniques to effectively assimilate XML, SQL, ADO.NET, and unstructured data sources. Build powerful LINQ queries, handle hierarchical and relational data, use lambdas and expression trees, and develop multi-threaded applications. Get proven strategies for handling conflicts and exceptions and creating custom extension methods. Additionally, you'll learn how to create your own LINQ provider through a working example, LINQ to Twitter.Filter, order, and group code using LINQ to ObjectsQuery relational and unstructured data with LINQ to SQL and LINQ to XML Work with ADO.NET through LINQ to DataSetDeploy object-oriented programming techniques and LINQ to Entities Read, manipulate, and create C#-based XML documentsGenerate DBML and external mapping files with SqlMetalCreate custom lambdas, expressions, providers, and extensionsSupport concurrent processing and multi-threading using PLINQBuild a custom LINQ provider for working with any data source&quot;" },
                new Book{ Id = 8,	CategoryName = categoryNames[0], ASIN = "0596159773", PublishDate = DateTime.Parse("2009-11-19"),  AddedDate = DateTime.Parse("2009-12-19"), Title = "jQuery Cookbook: Solutions & Examples for jQuery Developers"                                                                     ,        Author = "Cody Lindley"               ,  Description = "jQuery simplifies building rich, interactive web frontends. Getting started with this JavaScript library is easy, but it can take years to fully realize its breadth and depth; this cookbook shortens the learning curve considerably. With these recipes, you'll learn patterns and practices from 19 leading developers who use jQuery for everything from integrating simple components into websites and applications to developing complex, high-performance user interfaces.   Ideal for newcomers and JavaScript veterans alike, jQuery Cookbook starts with the basics and then moves to practical use cases with tested solutions to common web development hurdles. You also get recipes on advanced topics, such as methods for applying jQuery to large projects.  Solve problems involving events, effects, dimensions, forms, themes, and user interface elements Learn how to enhance your forms, and how to position and reposition elements on a page Make the most of jQuery's event management system, including custom events and custom event data Create UI elements-such as tabs, accordions, and modals-from scratch Optimize your code to eliminate bottlenecks and ensure peak performance Learn how to test your jQuery applications The book's contributors include:  Cody Lindley James Padolsey Ralph Whitbeck Jonathan Sharp Michael Geary Scott GonzÃ¡lez Rebecca Murphey Remy Sharp Ariel Flesler Brian Cherne JÃ¶rn Zaefferer Mike Hostetler Nathan Smith Richard D. Worth Maggie Wachs, Scott Jehl, Todd Parker, and Patty Toland Rob Burns&quot;" },
                new Book{ Id = 9,	CategoryName = categoryNames[0], ASIN = "1933988355", PublishDate = DateTime.Parse("2008-02-07"),  AddedDate = DateTime.Parse("2008-03-07"), Title = "jQuery in Action"                                                                                                                ,        Author = "Bear Bibeault"              ,  Description = "A good web development framework anticipates what you need to do and makes those tasks easier and more efficient; jQuery practically reads your mind. Developers of every stripe-hobbyists and professionals alike-fall in love with jQuery the minute they've reduced 20 lines of clunky JavaScript into three lines of elegant, readable code. This new, concise JavaScript library radically simplifies how you traverse HTML documents, handle events, perform animations, and add Ajax interactions to your web pages.  jQuery in Action, like jQuery itself, is a concise tool designed to make you a more efficient and effective web developer. In a short 300 pages, this book introduces you to the jQuery programming model and guides you through the major features and techniques you'll need to be productive immediately.  The book anchors each new concept in the tasks you'll tackle in day-to-day web development and offers unique lab pages where you immediately put your jQuery knowledge to work.  There are dozens of JavaScript libraries available now, with major companies like Google, Yahoo and AOL open-sourcing their in-house tools. This book shows you how jQuery stacks up against other libraries and helps you navigate interaction with other tools and frameworks.  jQuery in Action offers a rich investigation of the up-and-coming jQuery library for client-side JavaScript. This book covers all major features and capabilities in a manner focused on getting the reader up and running with jQuery from the very first sections. Web Developers reading this book will gain a deep understanding of how to use jQuery to simplify their pages and lives, as well as learn the philosophy behind writing jQuery-enhanced pages.&quot;" },
                new Book{ Id = 10,	CategoryName = categoryNames[0], ASIN = "1430210079", PublishDate = DateTime.Parse("2009-04-30"),  AddedDate = DateTime.Parse("2009-05-30"), Title = "Pro ASP.NET MVC Framework"                                                                                                       ,        Author = "Steven Sanderson"           ,  Description = "Steven Sanderson has seen the ASP.NET MVC framework mature from the start, so his experience, combined with comprehensive coverage of all the new features, including those in the official MVC development toolkit, offers the clearest understanding of how this exciting new framework could improve your coding efficiency and you'll gain invaluable up-to-date awareness of security, deployment, and interoperability challenges. The ASP.NET MVC Framework is the latest evolution of Microsoft's ASP.NET web platform. It introduces a radically new highâ€“productivity programming model that promotes cleaner code architecture, test-driven development, and powerful extensibility, combined with all the benefits of ASP.NET 3.5. An integral benefit of this book is that the core Model-View-Controller architectural concepts are not simply explained or discussed in isolation, but demonstrated in action. You'll work through an extended tutorial to create a working eâ€“commerce web application that combines ASP.NET MVC with the latest C# 3.0 language features and unit-testing best practices. By gaining this invaluable, practical experience, you can discover MVCs strengths and weaknesses for yourself and put your best learned theory into practice.&quot;" },
                new Book{ Id = 11,	CategoryName = categoryNames[0], ASIN = "0470384611", PublishDate = DateTime.Parse("2009-04-01"),  AddedDate = DateTime.Parse("2009-05-01"), Title = "Professional ASP.NET MVC 1.0 (Wrox Programmer to Programmer)"                                                                    ,        Author = "Scott Guthrie"              ,  Description = "This book begins with you working along as Scott Guthrie builds a complete ASP.NET MVC reference application. He begins NerdDinner by using the File->New Project menu command within Visual Studio to create a new ASP.NET MVC Application. You'll then incrementally add functionality and features. Along the way you'll cover how to create a database, build a model layer with business rule validations, implement listing/details data browsing, provide CRUD (Create, Update, Delete) data form entry support, implement efficient data paging, reuse UI using master pages and partials, secure the application using authentication and authorization, use AJAX to deliver dynamic updates and interactive map support, and implement automated unit testing. From there, the bulk of the rest of the book begins with the basic concepts around the model view controller pattern, including the little history and the state of the MVC on the web today.&quot;" },
                new Book{ Id = 12,	CategoryName = categoryNames[0], ASIN = "0672330318", PublishDate = DateTime.Parse("2010-02-28"),  AddedDate = DateTime.Parse("2010-03-01"), Title = "The Zen of XAML: Designer and Developer Collaboration in WPF and Silverlight"                                                    ,        Author = "Robby Ingebretsen"          ,  Description = "With XAML, Silverlight, and the Expression suite of web development software, Microsoft has provided all the tools that designers, developers, and managers need to collaborate on creating breakthrough user experiences. However, designers and developers have traditionally found it difficult to collaborate, and the mere presence of the tools doesn't solve that problem. There's been a growing clamor for reliable, practical information on making this collaboration work in real enterprises. In The Zen of XAML, Robby Ingebretsen offers this desperately needed guidance. Ingrebetsen is singularly well qualified to write this book. He is a leading edge &quot;creative developer&quot; with a deep understanding both the technical and design issues involved - and also former program manager for Windows Presentation Foundation (WPF). Ingrebetsen offers a specific and detailed roadmap for effective collaboration and workflow on projects that involve both state-of-the-art design and software development. Drawing on his unique experience, he shares a nuanced perspective on the collaboration-enabling capabilities of WPF and Silverlight. He gives designers the information and support they need to buy in and &quot;take ownership&quot; of their new integration-related roles; and helps developers understand how implementing workflow can benefit them, and what they need to do to support it.&quot;" },
                new Book{ Id = 13,	CategoryName = categoryNames[0], ASIN = "0596526733", PublishDate = DateTime.Parse("2006-03-27"),  AddedDate = DateTime.Parse("2006-04-27"), Title = "XAML in a Nutshell"                                                                                                              ,        Author = "Lori A. MacVittie"          ,  Description = "When Microsoft releases Windows Vista, the new operating system will support applications that employ graphics now used by computer games--clear, stunning, and active. The cornerstone for building these new user interfaces is XAML (&quot;Zammel&quot;), the XML-based markup language that works with Windows Presentation Foundation (WPF), Vista's new graphics subsystem.  An acronym for Extensible Application Markup Language, XAML offers a wealth of new controls and elements with exciting capabilities, including animation and rendering of 3D graphics. Windows developers are already jazzed by the possibilities of using XAML for fixed and flow format documents like PDF and HTML, 2D and 3D vector-based graphics, form development, animation, audio and video, transparent layering, and a lot more. Many feel that XAML will eliminate the need for multiple file formats or plug-ins (read: Flash), while lowering development costs and reducing time to market.  The problem is, most developers don't know XAML. While it is fairly easy to understand, you still need a quick guide to bring you up to speed before Vista's release, and that's where this book's simple, no nonsense approach comes in.  &quot;XAML in a Nutshell&quot; covers everything necessary to design user interfaces and .NET applications that take advantage of WPF. Prerequisites such as Microsoft's new unified build system, MSBuild, and core XAML constructs and syntax--including shortcuts--are all presented with plenty of examples to get you started. The Core XAML Reference section lets you dig even deeper into syntax rules and attributes for all XAML elements with a series of quick-reference chapters. This section divides XAML elements into logicalcategories of elements, controls, shapes and geometry, layout, animations, and transformations for easy reference.  &quot;XAML in a Nutshell&quot; helps you learn, firsthand, how to use this XML-based markup language to implement the new generation of user interface graphics. As one reviewer noted, &quot;Strong code examples and an efficient, conversational style take the tedium out of learning XAML and make the subject understandable--even interesting.&quot;&quot;" },
                new Book{ Id = 14,	CategoryName = categoryNames[0], ASIN = "1590598148", PublishDate = DateTime.Parse("2007-03-26"),  AddedDate = DateTime.Parse("2007-04-26"), Title = "Microformats: Empowering Your Markup for Web 2.0"                                                                                ,        Author = "John Allsopp"               ,  Description = "Microformats burst onto the scene a couple of years ago and are fast becoming an essential tool for all professional web designers and developers. Imagine being able to integrate all of your web-based contact details, tagged articles, and geographical information seamlessly in web and desktop applications, without having to add anything extra to your websites except a little specialized HTML markup. Microformats provide a more formalized technology for adding commonly used semantics (such as contact details, location, and reviews) to today's Web. Unlike XML or the semantic Web, microformats use ubiquitous technologies like HTML and XHTML, existing developer skills, and current web tools, and, perhaps most important, they work in all of today's web browsers.        This book is a comprehensive guide to microformats. It explores why, in Bill Gates's words, &quot;We need microformats&quot;; how microformats work; and the kinds of problems microformats help solve. the book covers every current microformat, with complete details of the syntax, semantics, and uses of each, along with real-world examples and a comprehensive survey of the tools available for working with them. the book also features case studies detailing how major web content publishers such as yahoo put microformats to work in their web applications. Written by one of the Web's best-known educators, John Allsopp, Microformats: Empowering Your Markup for Web 2.0 will help you painlessly get up to speed with this exciting technology.  In this book you'll learn:  How to take best advantage of the built-in semantics of XHTML and HTML  How to extend the semantics of HTML using microformats and open up a world of new possibilities with web applications  Every aspect of all the common microformats currently in use    How microformats help your websites and applications easily integrate with web applications like Google Maps, as well as desktop applications like iCal, outlook, and entourage  What innovative publishers and services, big and small, are doing right now with microformats&quot;" },
                new Book{ Id = 15,	CategoryName = categoryNames[1], ASIN = "0470458429", PublishDate = DateTime.Parse("2009-02-17"),  AddedDate = DateTime.Parse("2009-03-17"), Title = "Twitter Power: How to Dominate Your Market One Tweet at a Time"                                                                  ,        Author = "Joel Comm"                  ,  Description = "Get the business leader's guide to using Twitter to gain competitive advantage. Since 2006, forward-thinking companies like Apple, JetBlue, Whole Foods, and GM have discovered the instant benefits of leveraging the social media phenomenon known as Twitter to reach consumers directly, build their brand, and increase sales. Twitter is at the leading edge of the social media movement, allowing members to connect with one another in real time via short text messages?called &quot;tweets&quot;?that can be received either via the Twitter site or by e-mail, instant messenger, or cell phone. Many companies have started building entire teams within their organization dedicated solely to responding to tweets from consumers about their brand. And this is just the beginning.        In Twitter Power, Internet marketing and Web innovation expert Joel Comm shows businesses and marketers how to integrate Twitter into their existing marketing strategies to build a loyal following among Twitter members, expand awareness for their product or service, and even handle negative publicity due to angry or disappointed consumers. The book also presents case studies of companies on the forefront of the Twitter movement, to help you develop your own social networking strategies. Twitter Power is the result of extensive testing and participation in the social networking community and is a must-have for any business that wants to keep up with the social media movement. Twitter Power features a foreword by Tony Robbins.&quot;" },
                new Book{ Id = 16,	CategoryName = categoryNames[1], ASIN = "0470529695", PublishDate = DateTime.Parse("2009-05-11"),  AddedDate = DateTime.Parse("2009-06-11"), Title = "Twitter Tips, Tricks, and Tweets"                                                                                                ,        Author = "Paul McFedries"             ,  Description = "A full-color, portable guide to getting all the fun and productivity from Twitter    Twitter is hot! It's used by everyone from teens keeping up with their friends to fundraising charities and organizations responding to natural disasters; even President Obama tweets. Twitter Tips, Tricks, and Tweets gets you in on the fun, taking you all the way from setting up an account to incorporating cool third-party applications.      Defined as microblogging, Twitter allows you to be as active or passive as you choose in keeping up with the conversation. Limited to 140 characters per comment, or &quot;tweet&quot;, it's designed for here-and-now communication. This handy guide gives you everything you need to know.&quot;" },
                new Book{ Id = 17,	CategoryName = categoryNames[1], ASIN = "0789738023", PublishDate = DateTime.Parse("2008-09-07"),  AddedDate = DateTime.Parse("2008-10-07"), Title = "Facebook Marketing: Leverage Social Media to Grow Your Business"                                                                 ,        Author = "Steve Holzner"              ,  Description = "Profit from Facebook!  High-Impact, Low-Cost Social Marketing That Works!  .  With more than 80,000,000 affluent, savvy members, Facebook is today's fastest-growing marketing opportunity! But traditional marketing methods won't work here. In Facebook Marketing, best-selling author Steven Holzner reveals new social marketing techniques that do work, and shows you exactly how to make the most of them. Using true case studies, Holzner introduces powerful new techniques from today's smartest Facebook marketers...and helps you avoid pitfalls that can cost you money and credibility. No matter what business you're in, you'll learn how to create bottom-up, â€œviralâ€ Facebook marketing programs that achieve maximum results at minimum cost!      Crafting your Facebook profile for maximum impact    Getting into the Facebook community: crucial dos and don'ts    Joining the right Facebook Groupsâ€”or starting your own    Creating a Facebook blog that attracts paying customers    Promoting products and services with free Facebook Marketplace classifieds    Hosting your own Facebook events: from company picnics to concerts    Successfully advertising on Facebook, without overspending    Promoting your business within today's most popular Facebook applications    Tracking the results of your advertising    Using brand-new viral video marketing techniques     Driving even more Web traffic to your Facebook pages    Building your own Facebook applications.&quot;" },
                new Book{ Id = 18,	CategoryName = categoryNames[1], ASIN = "0470411554", PublishDate = DateTime.Parse("2009-05-04"),  AddedDate = DateTime.Parse("2009-06-04"), Title = "The Social Media Bible: Tactics, Tools, and Strategies for Business Success"                                                     ,        Author = "Lon Safko"                  ,  Description = "The Ultimate Guide To Social Media Marketing. The Social Media Bible will show you how to build or transform your business into a social media enabled enterprise where customers, employees, and prospects connect, collaborate, and champion your products, your services, and your way of doing business. Using the systematic approach presented in this comprehensive guide, you'll learn how to:  Increase your company and brand value by engaging people in new forms of communication, collaboration, education, and entertainment   Determine which social media tactics you should be using with your customers and employees            Evaluate and categorize the tools and applications that constitute the rapidly evolving social media ecosystem   Make social media tools like Facebook, MySpace, YouTube, Twitter, blogging, podcasting, and hundreds of others a part of your business strategy  Do a social media analysis inside your company to improve internal operations and outside your company to create and monetize relationships with customers and prospects  Implement social media micro- and macrostrategies to give your business the competitive edge it needs to survive and thrive  Virtually every business can use social media to increase sales and profits, and The Social Media Bible will show you how.  Part One introduces you to social media, and gives you a helpful framework, and presents practical and tactical tips for using some of these tools. Part Two introduces you to over 100 social media tools and applications in fifteen different categories, giving you a quick rundown of the features and functions of the tools that should become part of your social media strategy. Part Three offers mini exercises and assessments to help you conduct a social media audit of your company, your competition, and your customers, so you can craft the perfect strategy for your business.  If you want to grow your business, you have to stay connected to your stakeholders- whether you're a big corporation, a small business, or even a nonprofit. The Social Media Bible will show you how to harness the collective wisdom and viral value of your stakeholders and stay ahead of your competition.    Www.TheSocialMediaBible.com&quot;" },
                new Book{ Id = 19,	CategoryName = categoryNames[1], ASIN = "1884956858", PublishDate = DateTime.Parse("2008-11-01"),  AddedDate = DateTime.Parse("2008-12-01"), Title = "Secrets of Social Media Marketing: How to Use Online Conversations and Customer Communities to Turbo-Charge Your Business!"      ,        Author = "Paul Gillin"                ,  Description = "Secrets of Social Media Marketing is a handbook for marketers and business owners to use in deciding how to employ the new social media for online marketing. Social media has quickly moved from the periphery of marketing into the forefront, but this is a new and quickly-evolving field and there are few established formulas for success. Building on the lessons set out in Gillin's acclaimed and oft-reviewed The New Influencers: A Marketer's Guide to the New Social Media, this book provides practical advice on strategy, tools, and tactics. It is a hands-on manual that will educate marketers on how to extend their brands, generate leads, and engage customer communities using online tools.&quot;" },
                new Book{ Id = 23,	CategoryName = categoryNames[0], ASIN = "0321534921", PublishDate = DateTime.Parse("2008-05-04"),  AddedDate = DateTime.Parse("2008-06-04"), Title = "Designing for the Social Web"                                                                                                    ,        Author = "Joshua Porter"              ,  Description = "No matter what type of web site or application you're building, social interaction among the people who use it will be key to its success. They will talk about it, invite their friends, complain, sing its high praises, and dissect it in countless ways. With the right design strategy you can use this social interaction to get people signing up, coming back regularly, and bringing others into the fold. With tons of examples from real-world interfaces and a touch of the underlying social psychology theory, Joshua Porter shows you how to design your next great social web application. Inside, you'll discover:â€¢ The real reasons why people participate online and the psychology behind themâ€¢ The Usage Lifecycleâ€”or how people use your web application over timeâ€¢ How to get people past that trickiest of hurdles: sign-upâ€¢ What to do when you've launched a web application and nobody is using itâ€¢ How to analyze the effectiveness of your application screens and flowsâ€¢ How to grow your social web application from zero users to 1000â€”and beyondDesigning for the social web is about much more than adding features. It's about embracing the social interaction of the people who make you successfulâ€”and then designing smartly to encourage it.  &quot;" },
                new Book{ Id = 24,	CategoryName = categoryNames[0], ASIN = "0596516258", PublishDate = DateTime.Parse("2009-01-19"),  AddedDate = DateTime.Parse("2009-02-19"), Title = "Designing Web Interfaces: Principles and Patterns for Rich Interactions"                                                         ,        Author = "Bill Scott"                 ,  Description = "With the recent advent of Ajax and the resurgence of Flash for developing web sites and applications, new patterns of interaction have emerged on the Web. In this book, Bill Scott provides insight on how to best take advantage of the power of these technologies for designing a great user experience through a series of best practices, summarized as eight key principles. Each principle and its nuances are illustrated in detail with real world examples and counter-examples from both inside and outside Yahoo! The design principles provide the rationale for how to apply a pattern. Design patterns provide a solution in context. The eight design principles are introduced as a set of principles focused on rich interaction, feedback and user data models. Benefits to reader: 1. Take-away the key principles for creating a rich experience on the web 2. Build a vocabulary around common patterns of interaction for a common language between engineering & design 3. Have numerous real-world examples to clearly understand the principles & patterns for future reference 4. Be able to apply the patterns & principles in real world design problems Includes a companion website: designingrichwebexperience.com&quot;" },
                new Book{ Id = 25,	CategoryName = categoryNames[0], ASIN = "0596522304", PublishDate = DateTime.Parse("2009-06-10"),  AddedDate = DateTime.Parse("2009-07-10"), Title = "Even Faster Web Sites: Performance Best Practices for Web Developers"                                                            ,        Author = "Steve Souders"              ,  Description = "Performance is critical to the success of any web site, and yet today's web applications push browsers to their limits with increasing amounts of rich content and heavy use of Ajax. In this book, Steve Souders, web performance evangelist at Google and former Chief Performance Yahoo!, provides valuable techniques to help you optimize your site's performance.  Souders' previous book, the bestselling High Performance Web Sites, shocked the web development world by revealing that 80% of the time it takes for a web page to load is on the client side. In Even Faster Web Sites, Souders and eight expert contributors provide best practices and pragmatic advice for improving your site's performance in three critical categories:     JavaScript--Get advice for understanding Ajax performance, writing efficient JavaScript, creating responsive applications, loading scripts without blocking other components, and more.  Network--Learn to share resources across multiple domains, reduce image size without loss of quality, and use chunked encoding to render pages faster.   Browser--Discover alternatives to iframes, how to simplify CSS selectors, and other techniques.     Speed is essential for today's rich media web sites and Web 2.0 applications. With this book, you'll learn how to shave precious seconds off your sites' load times and make them respond even faster.  This book contains six guest chapters contributed by Dion Almaer, Doug Crockford, Ben Galbraith, Tony Gentilcore, Dylan Schiemann, Stoyan Stefanov, Nicole Sullivan, and Nicholas C. Zakas.&quot;" },
                new Book{ Id = 26,	CategoryName = categoryNames[0], ASIN = "0596102356", PublishDate = DateTime.Parse("2006-05-16"),  AddedDate = DateTime.Parse("2006-06-16"), Title = "Building Scalable Web Sites: Building, Scaling, and Optimizing the Next Generation of Web Applications"                          ,        Author = "Cal Henderson"              ,  Description = "Learn the tricks of the trade so you can build and architect applications that scale quickly--without all the high-priced headaches and service-level agreements associated with enterprise app servers and proprietary programming and database products. Culled from the experience of the Flickr.com lead developer, Building Scalable Web Sites offers techniques for creating fast sites that your visitors will find a pleasure to use.  Creating popular sites requires much more than fast hardware with lots of memory and hard drive space. It requires thinking about how to grow over time, how to make the same resources accessible to audiences with different expectations, and how to have a team of developers work on a site without creating new problems for visitors and for each other.  Presenting information to visitors from all over the world   Integrating email with your web applications   Planning hardware purchases and hosting options to have as much as you need without breaking your wallet   Partitioning and distributing databases to support large datasets and simultaneous transactions   Monitoring your applications to find and clear bottlenecks  * Providing services APIs and using services from other providers to increase your site's reach and capabilities  Whether you're starting a small web site with hopes of growing big or you already have a large system that needs maintenance, you'll find Building Scalable Web Sites to be a library of ideas for making things work.&quot;" },
                new Book{ Id = 27,	CategoryName = categoryNames[0], ASIN = "0596529309", PublishDate = DateTime.Parse("2007-09-11"),  AddedDate = DateTime.Parse("2007-10-11"), Title = "High Performance Web Sites: Essential Knowledge for Front-End Engineers"                                                         ,        Author = "Steve Souders"              ,  Description = "Want your web site to display more quickly? This book presents 14 specific rules that will cut 25% to 50% off response time when users request a page. Author Steve Souders, in his job as Chief Performance Yahoo!, collected these best practices while optimizing some of the most-visited pages on the Web. Even sites that had already been highly optimized, such as Yahoo! Search and the Yahoo! Front Page, were able to benefit from these surprisingly simple performance guidelines.  The rules in High Performance Web Sites explain how you can optimize the performance of the Ajax, CSS, JavaScript, Flash, and images that you've already built into your site -- adjustments that are critical for any rich web application. Other sources of information pay a lot of attention to tuning web servers, databases, and hardware, but the bulk of display time is taken up on the browser side and by the communication between server and browser. High Performance Web Sites covers every aspect of that process.  Each performance rule is supported by specific examples, and code snippets are available on the book's companion web site. The rules include how to:  Make Fewer HTTP Requests Use a Content Delivery Network  Add an Expires Header  Gzip Components  Put Stylesheets at the Top  Put Scripts at the Bottom  Avoid CSS Expressions  Make JavaScript and CSS External  Reduce DNS Lookups  Minify JavaScript  Avoid Redirects  Remove Duplicates Scripts  Configure ETags  Make Ajax Cacheable   If you're building pages for high traffic destinations and want to optimize the experience of users visiting your site, this book is indispensable.  &quot;If everyone would implement just 20% of Steve's guidelines, the Web would be a dramatically better place. Between this book and Steve's YSlow extension, there's really no excuse for having a sluggish web site anymore.&quot; -Joe Hewitt, Developer of Firebug debugger and Mozilla's DOM Inspector &quot;Steve Souders has done a fantastic job of distilling a massive, semi-arcane art down to a set of concise, actionable, pragmatic engineering steps that will change the world of web performance.&quot; -Eric Lawrence, Developer of the Fiddler Web Debugger, Microsoft Corporation&quot;" },
                new Book{ Id = 28,	CategoryName = categoryNames[0], ASIN = "0596517742", PublishDate = DateTime.Parse("2008-12-01"),  AddedDate = DateTime.Parse("2009-01-01"), Title = "JavaScript: The Good Parts"                                                                                                      ,        Author = "Douglas Crockford"          ,  Description = "Most programming languages contain good and bad parts, but JavaScript has more than its share of the bad, having been developed and released in a hurry before it could be refined. This authoritative book scrapes away these bad features to reveal a subset of JavaScript that's more reliable, readable, and maintainable than the language as a whole-a subset you can use to create truly extensible and efficient code.  Considered the JavaScript expert by many people in the development community, author Douglas Crockford identifies the abundance of good ideas that make JavaScript an outstanding object-oriented programming language-ideas such as functions, loose typing, dynamic objects, and an expressive object literal notation. Unfortunately, these good ideas are mixed in with bad and downright awful ideas, like a programming model based on global variables.   When Java applets failed, JavaScript became the language of the Web by default, making its popularity almost completely independent of its qualities as a programming language. In JavaScript: The Good Parts, Crockford finally digs through the steaming pile of good intentions and blunders to give you a detailed look at all the genuinely elegant parts of JavaScript, including:   Syntax Objects Functions Inheritance Arrays Regular expressions Methods Style Beautiful features The real beauty? As you move ahead with the subset of JavaScript that this book presents, you'll also sidestep the need to unlearn all the bad parts. Of course, if you want to find out more about the bad parts and how to use them badly, simply consult any other JavaScript book.   With JavaScript: The Good Parts, you'll discover a beautiful, elegant, lightweight and highly expressive language that lets you create effective code, whether you're managing object libraries or just trying to get Ajax to run fast. If you develop sites or applications for the Web, this book is an absolute must.&quot;" },
                new Book{ Id = 29,	CategoryName = categoryNames[0], ASIN = "1847194141", PublishDate = DateTime.Parse("2008-07-24"),  AddedDate = DateTime.Parse("2008-08-24"), Title = "Object-Oriented JavaScript: Create scalable, reusable high-quality JavaScript applications and libraries"                        ,        Author = "Stoyan Stefanov"            ,  Description = "Once listed in the &quot;nice to have&quot; sections of job postings, these days the knowledge of JavaScript is a deciding factor when it comes to hiring web developers. And rightly so. Where in the past we used to have the occasional few lines of JavaScript embedded in a web page, now we have advanced libraries and extensible architectures, powering the &quot;fat-client&quot;, AJAX-type rich internet applications.    JavaScript is the language of the browser, but it's also heavily employed in many other environments: server-side programming, desktop applications, application extensions and widgets. It's a pretty good deal: you learn one language and then code all kinds of different applications. While this book has one chapter specifically dedicated to the web browser environment including DOM, events, and AJAX tutorials, the rest is applicable to all the other environments too.     This book treats JavaScript as a serious object-oriented language, showing you how to build robust, maintainable, and powerful libraries and applications. Along the way, we cover many of the recent innovations such as AJAX, JSON, and interesting design and coding patterns. After reading this book, you'll be prepared to ace your JavaScript job interview and even impress with some bits that the interviewer maybe didn't know. You should read this book if you want to be able to take your JavaScript skills to a new level of sophistication.&quot;" },
                new Book{ Id = 30,	CategoryName = categoryNames[0], ASIN = "159059908X", PublishDate = DateTime.Parse("2007-12-10"),  AddedDate = DateTime.Parse("2008-01-10"), Title = "Pro JavaScript Design Patterns (Recipes: a Problem-Solution Ap)"                                                                 ,        Author = "Ross Harmes"                ,  Description = "As a web developer, you'll already know that JavaScript is a powerful language, allowing you to add an impressive array of dynamic functionality to otherwise static web sites. But there is more power waiting to be unlockedâ€”JavaScript is capable of full objectâ€“oriented capabilities, and by applying OOP principles, best practices, and design patterns to your code, you can make it more powerful, more efficient, and easier to work with alone or as part of a team.        With Pro JavaScript Design Patterns, you'll start with the basics of object-oriented programming in JavaScript applicable to design patterns, including making JavaScript more expressive, inheritance, encapsulation, information hiding, and more. With that covered, you can kick-start your JavaScript development in the second part of the book, where you'll find detail on how to implement and take advantage of several design patterns in JavaScript, including composites, decorators, facades, adapters, and many more.&quot;" },
                new Book{ Id = 31,	CategoryName = categoryNames[0], ASIN = "1590597273", PublishDate = DateTime.Parse("2006-12-11"),  AddedDate = DateTime.Parse("2007-01-11"), Title = "Pro JavaScript Techniques"                                                                                                       ,        Author = "John Resig"                 ,  Description = "Pro JavaScript Techniques is the ultimate JavaScript book for the modern web developer. It provides everything you need to know about modern JavaScript, and shows what JavaScript can do for your web sites. This book doesnt waste any time looking at things you already know, like basic syntax and structures.     Expert web developer and author John Resig concentrates on fundamental, vital topicswhat modern JavaScripting is (and isnt), the current state of browser support, and pitfalls to be wary of. The book is organized into four sections:      Modern JavaScript developmentusing JavaScript the object-oriented way, creating reusable code, plus testing and debugging  DOM scriptingupdating content and styles, plus events, and effect and event libraries  Ajaxhow Ajax works, overcoming problems, and using libraries to speed up development of Ajax applications  The future of JavaScriptlooking at cutting edge topics like JSON, HTML 5, and more      All concepts are backed up by real-world examples and case studies, and John provides numerous reusable functions and classes to save you time in your development. There are also up-to-date reference appendixes for the DOM, events, browser support (including IE7), and frameworksso you can look up specific details quickly and easily.&quot;" },
                new Book{ Id = 32,	CategoryName = categoryNames[0], ASIN = "1590598040", PublishDate = DateTime.Parse("2007-04-23"),  AddedDate = DateTime.Parse("2007-05-23"), Title = "Pro CSS and HTML Design Patterns"                                                                                                ,        Author = "Michael Bowers"             ,  Description = "If you are looking for a complete site design solutions book, this is not it, but if you are seeking a book to help you solve many common problemsâ€”then this book will serve as a useful tool for you. Pro CSS and HTML Design Patterns is a reference book and a cookbook on how to style web pages using CSS and XHTML. It contains 350 ready to use patterns (CSS and XHTML code snippets) you can copy and paste into your code. Each pattern can be combined with other patterns to create an unlimited number of solutions. Each pattern works reliably in all major browsers without the need for browser hacks.  You can interact with all the examples in the book at www.cssDesignPatterns.com.&quot;" },
                new Book{ Id = 33,	CategoryName = categoryNames[1], ASIN = "0596522347", PublishDate = DateTime.Parse("2008-08-12"),  AddedDate = DateTime.Parse("2008-09-12"), Title = "slide:ology: The Art and Science of Creating Great Presentations"                                                                ,        Author = "Nancy Duarte"               ,  Description = "No matter where you are on the organizational ladder, the odds are high that you've delivered a high-stakes presentation to your peers, your boss, your customers, or the general public. Presentation software is one of the few tools that requires professionals to think visually on an almost daily basis. But unlike verbal skills, effective visual expression is not easy, natural, or actively taught in schools or business training programs. slide:ology fills that void.  Written by Nancy Duarte, President and CEO of Duarte Design, the firm that created the presentation for Al Gore's Oscar-winning film, An Inconvenient Truth, this book is full of practical approaches to visual story development that can be applied by anyone. The book combines conceptual thinking and inspirational design, with insightful case studies from the world's leading brands. With slide:ology you'll learn to:   Connect with specific audiences Turn ideas into informative graphics Use sketching and diagramming techniques effectively Create graphics that enable audiences to process information easily Develop truly influential presentations Utilize presentation technology to your advantage Millions of presentations and billions of slides have been produced -- and most of them miss the mark. slide:ology will challenge your traditional approach to creating slides by teaching you how to be a visual thinker. And it will help your career by creating momentum for your cause.&quot;" },
                new Book{ Id = 34,	CategoryName = categoryNames[1], ASIN = "0321525655", PublishDate = DateTime.Parse("2008-01-04"),  AddedDate = DateTime.Parse("2008-02-04"), Title = "Presentation Zen: Simple Ideas on Presentation Design and Delivery"                                                              ,        Author = "Garr Reynolds"              ,  Description = "Presentation designer and internationally acclaimed communications expert Garr Reynolds, creator of the most popular Web site on presentation design and delivery on the net - presentationzen.com - shares his experience in a provocative mix of illumination, inspiration, education, and guidance that will change the way you think about making presentations with PowerPoint or Keynote. Presentation Zen challenges the conventional wisdom of making &quot;slide presentations&quot; in today-s world and encourages you to think differently and more creatively about the preparation, design, and delivery of your presentations. Garr shares lessons and perspectives that draw upon practical advice from the fields of communication and business. Combining solid principles of design with the tenets of Zen simplicity, this book will help you along the path to simpler, more effective presentations.&quot;" },
                new Book{ Id = 35,	CategoryName = categoryNames[1], ASIN = "0321668790", PublishDate = DateTime.Parse("2009-12-28"),  AddedDate = DateTime.Parse("2010-01-28"), Title = "Presentation Zen Design: Simple Design Principles and Techniques to Enhance Your Presentations"                                  ,        Author = "Garr Reynolds"              ,  Description = "In his internationally acclaimed, best-selling book Presentation Zen: Simple Ideas on Presentation Design and Delivery, presentation master Garr Reynolds gave readers the framework for planning, putting together, and delivering successful presentations. Now, he takes us further into the design realm and shows how we can apply time-honored design principles to presentation layouts.Throughout Presentation Zen Design, Garr shares his lessons on designing effective presentations that contain text, graphs, color, images, and video. After establishing guidelines for each of the various elements, he explains how to achieve an overall harmony and balance using the tenets of Zen simplicity. Not only will you discover how to design your slides for more professional-looking presentations, you'll learn to communicate more clearly and will accomplish the goal of making a stronger, more lasting connection with your audience.&quot;" },
                new Book{ Id = 36,	CategoryName = categoryNames[0], ASIN = "020161622X", PublishDate = DateTime.Parse("1999-10-30"),  AddedDate = DateTime.Parse("1999-11-30"), Title = "The Pragmatic Programmer: From Journeyman to Master"                                                                             ,        Author = "Andrew Hunt"                ,  Description = "If I'm putting together a project, it's the  authors of this book that I want. . . . And failing that I'd settle  for people who've read their book.&quot; -- Ward  Cunningham  Straight from the programming trenches, The Pragmatic Programmer  cuts through the increasing specialization and technicalities of  modern software development to examine the core process--taking a  requirement and producing working, maintainable code that delights its  users. It covers topics ranging from personal responsibility and  career development to architectural techniques for keeping your code  flexible and easy to adapt and reuse. Read this book, and you'll learn  how to    Fight software rot;  Avoid the trap of duplicating knowledge;  Write flexible, dynamic, and adaptable code;  Avoid programming by coincidence;  Bullet-proof your code with contracts, assertions, and exceptions;  Capture real requirements;  Test ruthlessly and effectively;  Delight your users;  Build teams of pragmatic programmers; and  Make your developments more precise with automation.    Written as a series of self-contained sections and filled with  entertaining anecdotes, thoughtful examples, and interesting  analogies, The Pragmatic Programmer illustrates the best practices and  major pitfalls of many different aspects of software  development. Whether you're a new coder, an experienced programmer, or  a manager responsible for software projects, use these lessons daily,  and you'll quickly see improvements in personal productivity,  accuracy, and job satisfaction. You'll learn skills and develop habits  and attitudes that form the foundation for long-term success in your  career. You'll become a Pragmatic Programmer.&quot;" },
                new Book{ Id = 37,	CategoryName = categoryNames[1], ASIN = "0321656210", PublishDate = DateTime.Parse("2009-10-25"),  AddedDate = DateTime.Parse("2009-11-25"), Title = "The Non-Designer's Presentation Book"                                                                                            ,        Author = "Robin Williams"             ,  Description = "These days, it's not just corporate marketing directors tasked with giving computer-based presentations anyone forced to stand in front of a crowd and talk for more than three minutes had better know how to put together a slide presentation. You're not a professional designer, but you want your slides to look professional. What do you do?.Enter Robin Williams, the beloved, best-selling non-designer's designer (with over 850,000 copies of The Non-Designer's Design Book in print!) who has taught an entire generation the basics of design and typography. In The Non-Designer's Presentation Book, Robin expands upon the design principles introduced in her award-winning Non-Designer's series. She explains four fundamental principles of good design as applied to digital presentations, and adds four more principles specific to clear communication with slides. Whether you work with a Mac or PC, PowerPoint or Keynote, let Robin guide you, in her signature, light-hearted style, through the entire process of creating a presentation from using the right software to organizing your ideas to designing effective, beautiful slides that won't put your audience to sleep.In this essential guide to presentation design, you'll learn:What makes a good presentation or a bad oneHow to plan, organize, and outline your presentationFour principles of designing effective presentationsFour principles for designing beautiful slides that communicate clearly.&quot;" },
                new Book{ Id = 38,	CategoryName = categoryNames[1], ASIN = "1591843065", PublishDate = DateTime.Parse("2009-12-31"),  AddedDate = DateTime.Parse("2010-01-31"), Title = "The Back of the Napkin (Expanded Edition): Solving Problems and Selling Ideas with Pictures"                                     ,        Author = "Dan Roam"                   ,  Description = "The acclaimed bestseller about visual problem solving-now bigger and better   &quot;There is no more powerful way to prove that we know something well than to draw a simple picture of it. And there is no more powerful way to see hidden solutions than to pick up a pen and draw out the pieces of our problem.&quot;  So writes Dan Roam in The Back of the Napkin, the international bestseller that proves that a simple drawing on a humble napkin can be more powerful than the slickest PowerPoint presentation. Drawing on twenty years of experience and the latest discoveries in vision science, Roam teaches readers how to clarify any problem or sell any idea using a simple set of tools.  He reveals that everyone is born with a talent for visual thinking, even those who swear they can't draw. And he shows how thinking with pictures can help you discover and develop new ideas, solve problems in unexpected ways, and dramatically improve your ability to share your insights.  Take Herb Kelleher and Rollin King, who figured out how to beat the traditional hub-and-spoke airlines with a bar napkin and a pen. Three dots to represent Dallas, Houston, and San Antonio. Three arrows to show direct flights. Problem solved, and the picture made it easy to sell Southwest Airlines to investors and customers.  Now with more color, bigger pictures, and additional content, this new edition does an even better job of helping you literally see the world in a new way. Join the teachers, project managers, doctors, engineers, assembly-line workers, pilots, football coaches, marine drill instructors, financial analysts, students, parents, and lawyers who have discovered the power of solving problems with pictures. &quot;" },
                new Book{ Id = 39,	CategoryName = categoryNames[1], ASIN = "1419620037", PublishDate = DateTime.Parse("2005-12-21"),  AddedDate = DateTime.Parse("2006-01-21"), Title = "7-Slide Solution(tm): Telling Your Business Story In 7 Slides or Less"                                                           ,        Author = "Paul J. Kelly"              ,  Description = "A unique approach to organizing and constructing business presentations that draws on the insights of cognitive psychology and provides an infrastructure to build presentations that resonate with your audience like a good story.&quot;" },
                new Book{ Id = 40,	CategoryName = categoryNames[3], ASIN = "0321385438", PublishDate = DateTime.Parse("2007-03-23"),  AddedDate = DateTime.Parse("2007-04-23"), Title = "The Adobe Photoshop Lightroom Book: The Complete Guide for Photographers"                                                        ,        Author = "Martin Evening"             ,  Description = "While Adobe Photoshop has long been their choice for editing digital  photographs, many photographers want a more focused tool. That's where  Adobe Photoshop Lightroom comes in. Designed from the ground up with  digital photographers in mind, Photoshop Lightroom offers powerful  editing features in a streamlined interface that lets photographers  import, sort, and organize images. The Adobe Photoshop Lightroom Book  was also written with photographers in mind. Best-selling author Martin  Evening describes Photoshop Lightroom's features in detail from a  photographer's perspective. As an established commercial and fashion  photographer, martin knows firsthand what photographers need for.  an efficient workflow. He's also been working with Photoshop Lightroom  from the beginning, monitoring the product's development and providing  feedback on the public beta. As a result, Martin knows the software  inside and out, from image selection to image editing to image  management. In this book you'll learn how to:Work efficiently with images shot in the raw formatn. import photographs with ease and sort them according to your workflowCreate and manage a personal image library Apply tonal corrections to multiple images quicklyIntegrate Photoshop Lightroom with Adobe Photoshop Export images for print or Web as digital contact sheets or personal portfoliosPhotographers  will find Adobe Photoshop Lightroomâ€”and The Adobe Photoshop Lightroom  Bookâ€”indispensable tools in their digital darkroom.  A free Lightroom 1.1 PDF supplement update is now available at www.peachpit.com/register.&quot;" },
                new Book{ Id = 41,	CategoryName = categoryNames[3], ASIN = "0321492161", PublishDate = DateTime.Parse("2007-04-16"),  AddedDate = DateTime.Parse("2007-05-16"), Title = "The Adobe Photoshop Lightroom Book for Digital Photographers"                                                                    ,        Author = "Scott Kelby"                ,  Description = "This book takes you beyond showing you  which sliders do what  to reveal the secrets of the new digital photography workflow using Adobe Lightroom, and he does it using three simple techniques that make this just a great learning tool: 1) Throughout the book Scott shares his own personal settings and studio tested techniques he s developed using Lightroom for his own photography workflow since well before Adobe released even the first Beta version. He knows what really works, what doesn t, and he tells you flat out which tools to use, which to avoid, and why. 2) The entire book is laid out in a real workflow order with everything step-by-step, so you can jump right in using Lightroom like a pro from the very start and sidestep a lot of productivity killing road blocks and time-wasting frustrations that might have tripped you up along the way. 3) In the last two bonus chapters Scott visually answers his No.1  most-asked  Lightroom question, which is:  Exactly what order am I supposed to do things in, and where does Photoshop fit in?  Scott teaches this by showing every step of the entire process, from the initial shoot to the final prints. Both chapters start with an on-location photo shoot, including full details on the equipment, camera settings, and even the lighting techniques. You ll see it all as he takes the photos from each shoot all the way through the entire workflow process, to the final output of the 16x20  prints for the client.&quot;" },
                new Book{ Id = 42,	CategoryName = categoryNames[3], ASIN = "0470400765", PublishDate = DateTime.Parse("2009-03-23"),  AddedDate = DateTime.Parse("2009-04-23"), Title = "Lightroom 2: Streamlining your Digital Photography Process"                                                                      ,        Author = "Nat Coalson"                ,  Description = "Adobe Photoshop Lightroom is the primary software-of-choice for today's digital photographer. As a new tool, using Lightroom requires mastering new skillsâ€”not always an easy task. Compounding the challenge, many photographers don't have fundamental knowledge of digital image processing. Lightroom 2: Streamlining your Digital Photography Process offers you. the digital imaging fundamentals and the Lightroom-specific skills you need to quickly learn and retain a mastery of Lightroom.        Written and designed by a professional photographer, designer, fine art printmaker and instructor, this book is designed to simplify and accelerate the learning process and is unique in both content and presentation. Nat Coalson is an Adobe Certified Expert in Lightroom and Photoshop, and has taught photography and imaging classes and workshops for many years. The approach taken by Lightroom 2: Streamlining your Digital Photography Process is based on over 20 years of professional experience in photography, printing, and imaging, and the methods of instruction were developed through the author's work with students and clients. At 400 pages, this book clearly explains how to use all of Lightroom's controls, and much more. Beginning with the original image capture, all the way through importing, editing, processing and presenting photos, the text also explains fundamental imaging concepts every modern photographer needs to know. Not just the &quot;how&quot; but also the &quot;why&quot;. Not just recipes, but core skills you will use again and again. And by developing stronger technical skills, you will be able to focus more on creative expression.        The content of Lightroom 2: Streamlining your Digital Photography Process also demonstrates a simple, practical workflow. You will learn real-world, professional-level production techniques through straightforward examples and easy-to-follow instructions.&quot;" },
                new Book{ Id = 43,	CategoryName = categoryNames[3], ASIN = "0240810937", PublishDate = DateTime.Parse("2009-09-14"),  AddedDate = DateTime.Parse("2009-10-14"), Title = "Digital Landscape Photography"                                                                                                   ,        Author = "John Gerlach"               ,  Description = "Photographing landscape with a film camera is different than with a digital camera. There are several books on the market that cover landscape photography but few of them are specifically for the digital photographer. This book is what you are looking for! Digital Landscape Photography covers:  * equipment such as accessories and lenses  * exposure from shutter speed.and other.common mistakes  * shooting  * light and its importance  * composing your perfect photo  * printing  * and a special section on specific subjects such as waterfalls and sunrises  Digital Landscape Photography, written by experts that have been shooting outdoors for decades, is a fresh look at current ways to shoot landscapes by making the most of digital format.  * the only photo book on landscape photography.in a landscape format  * both beautifully illustrated and instructive at the same time  * Authors are well known photographers in the field  Pictures have been published in National Wildlife, Sierra, Natural History, Outdoor Photographer, and many other major publications&quot;" },
                new Book{ Id = 44,	CategoryName = categoryNames[3], ASIN = "0240808568", PublishDate = DateTime.Parse("2007-07-25"),  AddedDate = DateTime.Parse("2007-08-25"), Title = "Digital Nature Photography: The Art and the Science"                                                                             ,        Author = "John Gerlach"               ,  Description = "1234 Over 50,000 photographers can't be wrong! John and Barbara Gerlach finally write the book their workshop attendees have been asking for!Digital Nature Photography is a how-to guide for photographers who want to take their work to the next level. Written by professionals with over 20 years experience, the Gerlachs reveal enlightening techniques for shooting nature images in the field. The combination of artistic approach and impeccable technique will help you capture your next great image. Learn what equipment works best in the field, and why, as well as tips for superb shots. Create sharp images at capture with detailed instructions for digital exposure. Master the three factors that lead to excellent images: technique, subject and situation&quot;" },
                new Book{ Id = 45,	CategoryName = categoryNames[3], ASIN = "1933952091", PublishDate = DateTime.Parse("2007-05-04"),  AddedDate = DateTime.Parse("2007-06-04"), Title = "Closeup Shooting: A Guide to Closeup, Tabletop and Macro Photography"                                                            ,        Author = "Cyrill Harnischmacher"      ,  Description = "Close-up photography is one of the most fascinating areas in photography. This illustrated guide will take the reader on a journey into the wonderful world of small, smaller, and smallest objects and show him how he can capture their beauty with photographic images. Each step of the way will be carefully explained; how to choose the right equipment, how to use ambient light or create artificial lighting, and how to conceptualize and frame the perfect shot.  Whereas the nature photographer is exploring facets and structures in his environment, the &quot;table top photographer&quot; is trying to shoot a small object, a product, or a small treasure for display on the web (e.g., eBay) or in print. Here, the choice of the appropriate lighting and backdrop, and the creative use of the camera's features are key to a perfect image.  Cyrill Harnischmacher explains all aspects of close-up shooting for both inside the studio, as well as outdoors. This book is filled with beautifully illustrated examples and detailed instructions on how to set up a system and workflow for successful close-up photography.&quot;" },
                new Book{ Id = 46,	CategoryName = categoryNames[3], ASIN = "1933952105", PublishDate = DateTime.Parse("2007-06-16"),  AddedDate = DateTime.Parse("2007-07-16"), Title = "Low Budget Shooting: Do It Yourself Solutions to Professional Photo Gear"                                                        ,        Author = "Cyrill Harnischmacher"      ,  Description = "The serious amateur photographer often faces the problem that even after all the dollars spent on camera, lenses, computer gear, and software, the spending never seems to end. More gear is needed for studio photography, tabletop photography, flash photography, and for accessories here and there. And in many cases, the right accessories are not even available. That is where this book comes in. Low Budget Shooting is the one-stop source where you will find instructions and a shopping list on how to build an array of useful and inexpensive photographic tools.  Filled with full-color images and easy-to-follow text, this book shows how to build essential lighting and studio equipment; how to make the perfect light-table for shooting small objects; and how to build reflectors, soft-boxes, and light-tents that really work. It also tells where to get some of the little helpers that make a photographer's life so much easier. This clever little book is a creative and valuable resource for most any photographer.&quot;" },
                new Book{ Id = 47,	CategoryName = categoryNames[3], ASIN = "1592008208", PublishDate = DateTime.Parse("2005-08-01"),  AddedDate = DateTime.Parse("2005-09-01"), Title = "Digital Food Photography"                                                                                                        ,        Author = "Lou Manna"                  ,  Description = "Great food always elicits strong reactions-the smell of freshly baked bread, the taste of a perfectly prepared steak. The job of a food photographer is to elicit that same mouth-watering reaction, but without the benefit of scent or taste. A well-shot photograph can send crowds flocking to a new restaurant or boost the sales of a culinary magazine. Capturing the perfect image requires a trained eye, finesse, and photographic skill. Digital Food Photography gives you the ingredients to cook up your own recipe for success-with professional lighting techniques, composition, food and prop styling, retouching, and tricks of the trade. You'll learn how digital photography combines teamwork, creativity, and technology, and how to make money creating delectable works of photographic art.&quot;" },
                new Book{ Id = 48,	CategoryName = categoryNames[3], ASIN = "1584282266", PublishDate = DateTime.Parse("2008-08-01"),  AddedDate = DateTime.Parse("2008-09-01"), Title = "Lighting for Photography: Techniques for Studio and Location Shoots"                                                             ,        Author = "Glenn Rand"                 ,  Description = "Presenting the tools.needed to create and control photographic lighting.with either digital.capture or film, this.book shows how to perfect lighting in the studio as well as on location. The tips in this resource cover the principles and rules of lighting that are not confined to any single pursuit, thus giving photographers an understanding of the rudiments of light and lighting, rather than a series of limited solutions that work only in certain situations. In addition to sample images, the book includes numerous diagrams and detailed captions that explain why and when particular lighting methods are chosen, as well as how they are executed. A technical book with an in-depth exploration of lighting, this resource will assist any photographer in understanding varying light conditions in both the studio and the natural world. &quot;" },
                new Book{ Id = 50,	CategoryName = categoryNames[2], ASIN = "1741046297", PublishDate = DateTime.Parse("2005-09-01"),  AddedDate = DateTime.Parse("2005-10-01"), Title = "The Travel Book: A Journey Through Every Country in the World"                                                                   ,        Author = "Lonely Planet Publications" ,  Description = "The world is a breathtakingly big place, and in this big book we have undertaken the big task of detailing as much of it as we can - every single country, many of the larger dependencies and other, smaller destinations. With the traveler's experience at its heart, this book shows a slice of life in every corner of the globe, and all points in between, engaging the reader's senses in an adventure which conjures up the sights, smells, tastes, sounds and feel of our amazing world.&quot;" },
                new Book{ Id = 51,	CategoryName = categoryNames[2], ASIN = "1426201257", PublishDate = DateTime.Parse("2007-10-16"),  AddedDate = DateTime.Parse("2007-11-16"), Title = "Journeys of a Lifetime: 500 of the World's Greatest Trips"                                                                       ,        Author = "National Geographic"        ,  Description = "No one knows the world like National Geographicâ€”and in this lavish volume, we reveal our picks for the world's most fabulous journeys, along with helpful information for readers who want to try them out. Compiled from the favorite trips of National Geographic's travel writers, Journeys of a Lifetime spans the globe to highlight the best of the world's most famous and lesser known sojourns. It presents an incredible diversity of possibilities, from ocean cruises around Antarctica to horse treks in the Andes. Every continent and every possible form of transport is covered. A timely resource for the burgeoning ranks of active travelers who crave adventurous and far-flung trips, Journeys of a Lifetime provides scores of creative ideas: trekking the heights of Mt. Kilimanjaro in Tanzania... mountain biking in Transylvania... driving through the scenic highlands of Scotland... or rolling through the outback on Australia's famous Ghan train... and dozens of other intriguing options all over the world. Journeys of a Lifetime also features 22 fun Top 10 lists in all sorts of categories. What are the world's top 10 elevator rides, bridges to walk across, trolley rides, ancient highways, or underground walking adventures? Readers will love evaluating and debating the selections. Each chapter showcases stunning photography, full-color maps, evocative text, and expert adviceâ€”including how to get there, when to visit, and how to make the most of the journeyâ€”all packaged in a luxurious oversize volume to treasure for years to come.&quot;" },
                new Book{ Id = 52,	CategoryName = categoryNames[2], ASIN = "0756653991", PublishDate = DateTime.Parse("2009-08-17"),  AddedDate = DateTime.Parse("2009-09-17"), Title = "Off the Tourist Trail: 1,000 Unexpected Travel Alternatives"                                                                     ,        Author = "DK Publishing"              ,  Description = "Off the Tourist Trail is a guide to the world's unspoiled sights and experiences. It takes a hundred clichÃ©d tourist destinations - everything from over-visited national parks to overrated museums - and reveal 1,000 fresh and fascinating alternative options. Written by a team of travel experts, and with a foreword by Bill Bryson, this book brings vibrant cities, enchanting sights, breathtaking natural wonders and unforgettable experiences to life with informative narrative and stunning photography.  Choose your destination by theme - Ancient and Historical Sights, Festivals and Parties, Great Journeys, Architectural Marvels, Natural Wonders, Beaches, Sports and Activities, Art and Culture, and Cities - or simply flick through this sumptuous guide and be inspired. Practical advice on getting there and around, where to stay, where to eat and when to go, as well as useful 'Need to Know' facts, ensure that you get the most out of your time away.  Less crowded, generally less expensive, and often more spectacular and rewarding, these lesser-known wonders of the world encourage readers to ditch the famous but well-worn choices, reminding them what real travel is all about - escaping the everyday and embracing the new.  Vacations will never be the same again.&quot;" },
                new Book{ Id = 53,	CategoryName = categoryNames[2], ASIN = "0470460296", PublishDate = DateTime.Parse("2009-09-08"),  AddedDate = DateTime.Parse("2009-10-08"), Title = "The Unofficial Guide to Las Vegas 2010"                                                                                          ,        Author = "Bob Sehlinger"              ,  Description = "      In 2008, Las Vegas hosted 37.5 million visitors according to the.Las Vegas Convention and Visitors Authority.      Global Travel.Industry News and Forbes Traveler rank Las Vegas as the #2 most popular U.S. destination to visit calling it &quot;America's favorite playground&quot;.      The Unofficial Guide to Las Vegas ranks over 100 hotels and casinos- the most offered by any other guidebook for the destination, providing complete detailed descriptions of each casino hotel.      &quot;" },
                new Book{ Id = 54,	CategoryName = categoryNames[2], ASIN = "0789495651", PublishDate = DateTime.Parse("2003-09-01"),  AddedDate = DateTime.Parse("2003-10-01"), Title = "Southwest USA & Las Vegas (Eyewitness Travel Guides)"                                                                            ,        Author = "DK Publishing"              ,  Description = "Covers: the Grand Canyon, Northern Arizona, Phoenix, Southern Arizona, Las Vegas, Southern Utah, the Four Corners, Santa Fe, Northern New Mexico, Albuquerque, and Southern New Mexico.&quot;" },
                new Book{ Id = 55,	CategoryName = categoryNames[2], ASIN = "0789496801", PublishDate = DateTime.Parse("2003-11-01"),  AddedDate = DateTime.Parse("2003-12-01"), Title = "Pacific Northwest (Eyewitness Travel Guides)"                                                                                    ,        Author = "Stephen Brewer"             ,  Description = "Highlights include: Astoria, Cannon Beach, Tillamook, Cape Mears, and other destinations in British Coulmia, Washington, and Oregon.&quot;" },
                new Book{ Id = 56,	CategoryName = categoryNames[2], ASIN = "0756645719", PublishDate = DateTime.Parse("2009-08-17"),  AddedDate = DateTime.Parse("2009-09-17"), Title = "Top 10 Seattle (Eyewitness Travel Guide)"                                                                                        ,        Author = "DK Publishing"              ,  Description = "Packed with the same reliable information and breathtaking color photography as our DK Eyewitness Travel Guides, DK Eyewitness Travel Top 10 Guides are fully illustrated, pocket-sized travel guides, with handy pull-out maps marking all the major sights.&quot;" }
            };

            books.ForEach(b => context.Books.Add(b));

            var flaggedBooks = new List<FlaggedBook>
            {
                new FlaggedBook { Id = 0, Profile = profiles[0], Book = books[51], Rating = 3,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 1, Profile = profiles[1], Book = books[0], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 4, Profile = profiles[0], Book = books[44], Rating = 2,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 7, Profile = profiles[1], Book = books[49], Rating = 5,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 8, Profile = profiles[1], Book = books[45], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 9, Profile = profiles[1], Book = books[43], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 10, Profile = profiles[5], Book = books[43], Rating = 1,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 11, Profile = profiles[5], Book = books[45], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 12, Profile = profiles[5], Book = books[49], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 13, Profile = profiles[2], Book = books[47], Rating = 2,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 14, Profile = profiles[2], Book = books[44], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 15, Profile = profiles[2], Book = books[50], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 16, Profile = profiles[0], Book = books[49], Rating = 2,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 19, Profile = profiles[0], Book = books[46], Rating = 2,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 21, Profile = profiles[0], Book = books[43], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 23, Profile = profiles[0], Book = books[35], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 27, Profile = profiles[0], Book = books[18], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 30, Profile = profiles[0], Book = books[36], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 31, Profile = profiles[0], Book = books[5], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 32, Profile = profiles[0], Book = books[4], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 35, Profile = profiles[0], Book = books[3], Rating = 3,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 37, Profile = profiles[0], Book = books[21], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 39, Profile = profiles[0], Book = books[29], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 41, Profile = profiles[0], Book = books[25], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 45, Profile = profiles[0], Book = books[45], Rating = 3,	IsFlaggedToRead = 0 },
                new FlaggedBook { Id = 46, Profile = profiles[0], Book = books[48], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 48, Profile = profiles[2], Book = books[5], Rating = 0,	IsFlaggedToRead = 1 },
                new FlaggedBook { Id = 49, Profile = profiles[0], Book = books[6], Rating = 0,	IsFlaggedToRead = 1 }
            };

            flaggedBooks.ForEach(fb => context.FlaggedBooks.Add(fb));
        }
    }

    public partial class Book
    {
        public Book()
        {
            this.FlaggedBooks = new HashSet<FlaggedBook>();
        }

        public int Id { get; set; }

        public int CategoryId { get; set; }

        public string ASIN { get; set; }

        public string Title { get; set; }

        public string Author { get; set; }

        public string Description { get; set; }

        public System.DateTime PublishDate { get; set; }

        public System.DateTime AddedDate { get; set; }

        [ForeignKey("CategoryId")]
        public virtual CategoryName CategoryName { get; set; }

        [Include]
        public virtual ICollection<FlaggedBook> FlaggedBooks { get; set; }
    }

    public partial class Category
    {
        public int Id { get; set; }

        public int ProfileId { get; set; }

        public int CategoryId { get; set; }

        [ForeignKey("CategoryId")]
        [Include]
        public virtual CategoryName CategoryName { get; set; }

        public virtual Profile Profile { get; set; }
    }

    public partial class CategoryName
    {
        public CategoryName()
        {
            this.Books = new HashSet<Book>();
            this.Categories = new HashSet<Category>();
        }

        public int Id { get; set; }

        public string Name { get; set; }

        public virtual ICollection<Book> Books { get; set; }

        public virtual ICollection<Category> Categories { get; set; }
    }

    public partial class FlaggedBook
    {
        public int Id { get; set; }

        public int ProfileId { get; set; }

        public int BookId { get; set; }

        public int Rating { get; set; }

        public int IsFlaggedToRead { get; set; }

        public virtual Book Book { get; set; }

        public virtual Profile Profile { get; set; }
    }

    public partial class Friend
    {
        [Key]
        public int Id { get; set; }

        public int ProfileId { get; set; }

        public int FriendId { get; set; }

        public virtual Profile Profile { get; set; }

        [ForeignKey("FriendId")]
        [Include]
        public virtual Profile FriendProfile { get; set; }
    }

    public partial class Profile
    {
        public Profile()
        {
            this.Categories = new HashSet<Category>();
            this.FlaggedBooks = new HashSet<FlaggedBook>();
            this.Friends = new HashSet<Friend>();
        }

        public int Id { get; set; }

        [Required]
        public string Name { get; set; }

        [Required, DataType(DataType.EmailAddress)]
        public string EmailAddress { get; set; }

        public string AspNetUserGuid { get; set; }

        [Include]
        public virtual ICollection<Category> Categories { get; set; }

        [Include]
        public virtual ICollection<FlaggedBook> FlaggedBooks { get; set; }

        [Include]
        public virtual ICollection<Friend> Friends { get; set; }
    }
}